--[[
	Addon: Raidificator
	Author: @Olivierko; patched by @code65536
	Version: 2.0.1

	Changelog:

	2.0.1
	* Bug Fix: Corrected a cosmetic typo

	2.0
	* New: Support for Maw of Lorkhaj
	* New: Support for the new vitality system; also, lives are now formatted as
	       "remaining/max" and will turn red when it reaches zero (to be
	       consistent with the default UI)
	* New: Changes and additions to the game's raid-related APIs have enabled
	       areas of substantial code simplification/cleanup and improvements to
	       reliability (e.g., the timer will no longer break after a crash)
	* New: Par completion time, from the new API; this time is now displayed in
	       a chatbox message at the start and end of a trial, and the time
	       display will turn red when the par time is exceeded (to be consistent
	       with the default UI)
	* Bug Fix: Fixed the nil results and the associated error popups
	* Misc: Updated icons
]]--

-------------------------------------------------------------------------------------------------
-- OnTrialStart event --
-------------------------------------------------------------------------------------------------
Raidificator.OnTrialStart = function(eventCode, trialName, isWeekly)

	Raidificator.InitializeTrial();
	Raidificator.Log(string.format("%s started; the par completion time is %d minutes.", trialName, Raidificator._state.targetTime / 60000));
end

-------------------------------------------------------------------------------------------------
-- OnTrialComplete event --
-------------------------------------------------------------------------------------------------
Raidificator.OnTrialComplete = function(eventCode, trialName, score, totalTime)

	-- totalTime is total ms
	local totalSeconds = totalTime / 1000;

	-- target time difference
	local tt = math.abs(Raidificator._state.targetTime - totalTime);
	local ttModifier = "outpaced";

	if (totalTime > Raidificator._state.targetTime) then
		ttModifier = "exceeded";
	end

	local ttMessage = string.format(
		"Your total time was %d:%02d:%02d.%03d; it %s the par completion time of %d minutes by %d:%02d minutes.",
		math.floor(totalSeconds / 3600),
		math.floor(totalSeconds / 60) % 60,
		math.floor(totalSeconds) % 60,
		totalTime % 1000,
		ttModifier,
		Raidificator._state.targetTime / 60000,
		math.floor(tt / 60000),
		math.floor(tt / 1000) % 60
	);

	local message = string.format(
		"%s completed with a final score of %d and with %d/%d vitality bonuses remaining.",
		trialName,
		score,
		Raidificator._state.revivesRemaining,
		Raidificator._state.revivesMax
	);

	-- log the result
	Raidificator.Log(message);
	Raidificator.Log(ttMessage);

	-- save to history
	Raidificator.Store(
		Raidificator._state.raidId,
		totalSeconds,
		score,
		Raidificator._state.revivesMax - Raidificator._state.revivesRemaining,
		Raidificator._state.revivesRemaining,
		Raidificator._playerName,
		Raidificator.GetGroupMembers()
	);

	-- update the main UI with the final results; calling UpdateGUI directly
	-- with the final figures, as the raid is no longer in progress
	Raidificator.UpdateGUI(totalTime, score);

	-- trial is over, reset
	Raidificator.ResetTrial();
end

-------------------------------------------------------------------------------------------------
-- OnTrialFailed event --
-------------------------------------------------------------------------------------------------
Raidificator.OnTrialFailed = function(eventCode, trialName, score)

	Raidificator.Log("Raid failed: " .. trialName);

	-- ensure that we update the main ui
	Raidificator.UpdateRaidInfo();

	-- trial is over, reset
	Raidificator.ResetTrial();
end

-------------------------------------------------------------------------------------------------
-- OnTrialReviveCounterUpdate event --
-------------------------------------------------------------------------------------------------
Raidificator.OnTrialReviveCounterUpdate = function(eventCode, counter)

	Raidificator._state.revivesRemaining = counter;
end

-------------------------------------------------------------------------------------------------
-- Reset trial --
-------------------------------------------------------------------------------------------------
Raidificator.ResetTrial = function()

	-- To avoid race conditions where a trial is reset before other calls that
	-- require this data, only the flags should be reset; the data itself can
	-- be reset lazily upon reinitialization.

	Raidificator._state.initialized = false;
	Raidificator._state.raidInProgress = false;
end

-------------------------------------------------------------------------------------------------
-- HideUI function --
-------------------------------------------------------------------------------------------------
Raidificator.HideUI = function()

	if Raidificator._state.forceShow then
		return
	end

	Raidificator._gui:SetHidden(true);
end

-------------------------------------------------------------------------------------------------
-- UpdateRaidInfo function --
-------------------------------------------------------------------------------------------------
Raidificator.UpdateRaidInfo = function()

	-- for some reason the ui isn't ready at start sometimes, guessing it's a timing issue
	if (Raidificator._gui == nil) then
		return
	end

	-- no ongoing raid
	if (not Raidificator._state.raidInProgress) then
		return
	end

	-- if for some reason the initialization had been missed, call it again
	if (not Raidificator._state.initialized) then
		Raidificator.InitializeTrial();
	end

	-- ensure display
	if (Raidificator._gui:IsHidden()) then
		Raidificator._gui:SetHidden(false);
	end

	Raidificator.UpdateGUI(GetRaidDuration(), GetCurrentRaidScore());
end

-------------------------------------------------------------------------------------------------
-- UpdateGUI function --
-------------------------------------------------------------------------------------------------
Raidificator.UpdateGUI = function(raidDurationMS, score)

	-- format the current raid time
	local raidDurationSec = math.floor(raidDurationMS / 1000);

	local raidTime = string.format(
		"%02d:%02d:%02d",
		math.floor(raidDurationSec / 3600),
		math.floor((raidDurationSec / 60) % 60),
		raidDurationSec % 60
	);

	Raidificator._gui.elapsedTime.label:SetText(raidTime);
	Raidificator._gui.score.label:SetText(score);
	Raidificator._gui.souls.label:SetText(Raidificator._state.revivesRemaining .. "/" .. Raidificator._state.revivesMax);

	-- Make the vitality display red if 0
	if (Raidificator._state.revivesRemaining > 0) then
		Raidificator._gui.souls.label:SetColor(1, 1, 1, 1);
	else
		Raidificator._gui.souls.label:SetColor(1, 0, 0, 1);
	end

	-- Make the time display red if the target time has been exceeded
	if (raidDurationMS <= Raidificator._state.targetTime) then
		Raidificator._gui.elapsedTime.label:SetColor(1, 1, 1, 1);
	else
		Raidificator._gui.elapsedTime.label:SetColor(1, 0, 0, 1);
	end
end

-------------------------------------------------------------------------------------------------
-- Poll function --
-------------------------------------------------------------------------------------------------
Raidificator.Poll = function()

	Raidificator._state.raidInProgress = IsRaidInProgress();

	if Raidificator._state.raidInProgress then
		Raidificator.UpdateRaidInfo();
	else
		-- reset any trial that's in progress
		Raidificator.ResetTrial();

		if (not Raidificator._gui:IsHidden()) then
			-- hide after 45 000 ms
			zo_callLater(Raidificator.HideUI, 45000);
		end
	end
end

-------------------------------------------------------------------------------------------------
-- Store function --
-------------------------------------------------------------------------------------------------
Raidificator.Store = function(id, timeInSeconds, score, deaths, revives, character, groupMembers)

	local now = GetTimeStamp();

	if Raidificator.SavedVariables.history == nil then
		Raidificator.SavedVariables.history = {};
	end

	if Raidificator.SavedVariables.history[id] == nil then
		Raidificator.SavedVariables.history[id] = {};
	end

	local trial = {
		id = id,
		timeInSeconds = timeInSeconds,
		score = score,
		deaths = deaths,
		revives = revives,
		character = character,
		completed = now,
		groupMembers = groupMembers
	};

	table.insert(Raidificator.SavedVariables.history[id], trial);
end

-------------------------------------------------------------------------------------------------
-- Initialize function --
-------------------------------------------------------------------------------------------------
Raidificator.Initialize = function()

	EVENT_MANAGER:UnregisterForEvent(Raidificator._tracker, EVENT_ADD_ON_LOADED);

	Raidificator.SavedVariables = ZO_SavedVars:NewAccountWide("RaidificatorSavedVariables", 1, nil, Raidificator._defaultSettings);

	SLASH_COMMANDS[Raidificator._slashCommand] = Raidificator.SlashCommands;

	-- build main ui
	Raidificator.BuildUI();

	-- register keybindings
	ZO_CreateStringId("SI_BINDING_NAME_TOGGLE_HISTORY", "Toggle History");

	-- initialize any ongoing raid
	Raidificator.InitializeTrial();
end

-------------------------------------------------------------------------------------------------
-- Initialize trial function --
-------------------------------------------------------------------------------------------------
Raidificator.InitializeTrial = function()

	Raidificator._state.raidInProgress = IsRaidInProgress();

	if Raidificator._state.raidInProgress then

		-- initialize state
		Raidificator._state.raidId = GetCurrentParticipatingRaidId();
		Raidificator._state.revivesRemaining = GetRaidReviveCountersRemaining();
		Raidificator._state.revivesMax = GetCurrentRaidStartingReviveCounters();
		Raidificator._state.targetTime = GetRaidTargetTime();
		Raidificator._state.initialized = true;
	end
end

-------------------------------------------------------------------------------------------------
-- GetGroupMembers function --
-------------------------------------------------------------------------------------------------
Raidificator.GetGroupMembers = function()

	local isInGroup = GetGroupSize() > 1;

	if not isInGroup then
		return {};
	end

	local groupMembers = {};

	local groupSize = GetGroupSize();

	for i = 1, groupSize do

		local unitTag = GetGroupUnitTagByIndex(i);

		local member = GetUnitName(unitTag);

		if member ~= Raidificator._playerName then
			table.insert(groupMembers, i, member);
		end
	end

	return groupMembers;
end

-------------------------------------------------------------------------------------------------
-- Regroup function --
-------------------------------------------------------------------------------------------------
Raidificator.Regroup = function()

	local isInGroup = GetGroupSize() > 1;

	if not isInGroup then
		Raidificator.Log("You're not in a group!");
		return
	end

	Raidificator.Log("Regrouping!");

	local groupMembers = Raidificator.GetGroupMembers();

	-- disband the group
	GroupDisband();

	-- regroup with fixed delay
	zo_callLater(function ()
		for key,value in pairs(groupMembers) do
			GroupInviteByName(value);
		end
	end, Raidificator._regroupDelay)
end

-------------------------------------------------------------------------------------------------
--  Toggles history window  --
-------------------------------------------------------------------------------------------------
Raidificator.ToggleHistory = function()

	local isHidden = Raidificator._gui.history:IsHidden();

	if (isHidden) then
		Raidificator._gui.history:SetHidden(false);
	else
		Raidificator._gui.history:SetHidden(true);
	end
end

-------------------------------------------------------------------------------------------------
--  Toggles UI window  --
-------------------------------------------------------------------------------------------------
Raidificator.Toggle = function()

	if Raidificator._state.forceShow then
		Raidificator._state.forceShow = false;
		Raidificator._gui:SetHidden(true);
	else
		Raidificator._state.forceShow = true;
		Raidificator._gui:SetHidden(false);
	end
end

-------------------------------------------------------------------------------------------------
--  Toggles the lock of the UI window --
-------------------------------------------------------------------------------------------------
Raidificator.Lock = function()

	if Raidificator.SavedVariables.isMovable then
		Raidificator.SavedVariables.isMovable = false;
		Raidificator._gui:SetMovable(false);
		Raidificator.Log("Main UI is now locked!");
	else
		Raidificator.SavedVariables.isMovable = true;
		Raidificator._gui:SetMovable(true);
		Raidificator.Log("Main UI is now movable!");
	end
end

-------------------------------------------------------------------------------------------------
--  Help function  --
-------------------------------------------------------------------------------------------------
Raidificator.Help = function()

	Raidificator.Log("|cFF3300//|r |cFFFF00      Raidificator 2.0      |r |cFF3300//|r");
	Raidificator.Log("|cFF3300|r|cFFFF00Available commands:|r |cFF3300|r");
	Raidificator.Log("|cFF3300-|r /rcr history|r |cFF3300 toggles the history ui (hotkey available)|r");
	Raidificator.Log("|cFF3300-|r /rcr regroup|r |cFF3300 regroups the current group|r");
	Raidificator.Log("|cFF3300-|r /rcr toggle|r |cFF3300 toggles the main ui|r");
	Raidificator.Log("|cFF3300-|r /rcr lock|r |cFF3300 toggles the lock of the main ui|r");
end

-------------------------------------------------------------------------------------------------
-- Log function --
-------------------------------------------------------------------------------------------------
Raidificator.Log = function(msg)
	d(msg);
end

-------------------------------------------------------------------------------------------------
--  Saves the GUI location  --
-------------------------------------------------------------------------------------------------
function Raidificator.SaveGUILocation()

	Raidificator.SavedVariables.offsetX = Raidificator._gui:GetLeft();
	Raidificator.SavedVariables.offsetY = Raidificator._gui:GetTop();
end

-------------------------------------------------------------------------------------------------
--  Initializes the slash commands  --
-------------------------------------------------------------------------------------------------
Raidificator.SlashCommands = function(arg)
	if (arg == Raidificator._slashCommands.regroup) then
		Raidificator.Regroup();
	elseif arg == Raidificator._slashCommands.history then
		Raidificator.ToggleHistory();
	elseif arg == Raidificator._slashCommands.toggle then
		Raidificator.Toggle();
	elseif arg == Raidificator._slashCommands.lock then
		Raidificator.Lock();
	elseif arg == Raidificator._slashCommands.help then
		Raidificator.Help();
	else
		Raidificator.Log("Unknown command, type '/rcr help' for list of commands.");
	end
end

-------------------------------------------------------------------------------------------------
-- OnAddOnLoaded --
-------------------------------------------------------------------------------------------------
function Raidificator.OnAddOnLoaded(event, addonName)

	if addonName ~= Raidificator.name then
		return
	end

	Raidificator:Initialize();
end

-------------------------------------------------------------------------------------------------
-- Register Events --
-------------------------------------------------------------------------------------------------
EVENT_MANAGER:RegisterForEvent(Raidificator._tracker, EVENT_ADD_ON_LOADED, Raidificator.Initialize);
EVENT_MANAGER:RegisterForEvent(Raidificator._tracker, EVENT_RAID_TRIAL_STARTED, Raidificator.OnTrialStart);
EVENT_MANAGER:RegisterForEvent(Raidificator._tracker, EVENT_RAID_TRIAL_COMPLETE, Raidificator.OnTrialComplete);
EVENT_MANAGER:RegisterForEvent(Raidificator._tracker, EVENT_RAID_TRIAL_FAILED, Raidificator.OnTrialFailed);
EVENT_MANAGER:RegisterForEvent(Raidificator._tracker, EVENT_RAID_REVIVE_COUNTER_UPDATE , Raidificator.OnTrialReviveCounterUpdate);
EVENT_MANAGER:RegisterForUpdate(Raidificator._tracker, Raidificator._pollInterval, Raidificator.Poll);
